#include "gpmain.h"
#include "shared.h"
#include "pcecrc.h"

/* Name of the loaded file */
char game_name[0x100];

/* split : 1= Split image (only needed for 512k versions of 384k images)
   flip  : 1= Bit-flip image (only for some TurboGrafx-16 images) */
int load_rom(char *filename, int split, int flip)
{

   #include "bitflip.h"
    uint8 header[0x200];
    uint8 *ptr = NULL, *buf = NULL;
    uint32 crc;
    int n; //size

	ERR_CODE err_code;
	F_HANDLE h_rfile;
	unsigned long n_dummy;
	unsigned long size;

	split = 0; flip = 0;

    /* Default */
	gp_str_func.strcpy(game_name, filename);

	//err_code = GpRelativePathSet("gp:\\gpmm\\pcegp");
	err_code = GpFileGetSize(filename, &size);
	err_code = GpFileOpen(filename, OPEN_R, &h_rfile);
	if (err_code != SM_OK)
	{
		textout((char*)"Error");
	}

	buf = (unsigned char*)gp_mem_func.malloc(size);	
	err_code = GpFileRead(h_rfile, buf, size, (ulong*)&n_dummy);
	if (err_code != SM_OK)
	{
		textout((char*)"Error");
	}

	GpFileClose(h_rfile);


//	gzFile *gd = NULL;

	/* Open file */
//	gd = gzopen(filename, "rb");
//	if(!gd) return (0);

	/* Get file size */
//	size = gzsize(gd);

	/* Allocate file data buffer */
/*	buf = malloc(size);
	if(!buf) {
		gzclose(gd);
		return (0);
	} */

	/* Read file data */
//	gzread(gd, buf, size);

	/* Close file */
//	gzclose(gd);

    /* Check for 512-byte header */
    ptr = buf;
    if((size / 512) & 1)
    {
        gp_str_func.memcpy(header, buf, 0x200);
        size -= 0x200;
        buf += 0x200;
    }

    // Generate CRC and print information
/*    crc = crc32(0, buf, size);

    // Look up game CRC in the CRC database, and set up flip and split options accordingly
    for(n = 0; n < (sizeof(pcecrc_list) / sizeof(t_pcecrc)); n += 1)
    {
        if(crc == pcecrc_list[n].crc)
        {
            if(pcecrc_list[n].flag & FLAG_BITFLIP) flip = 1;
            if(pcecrc_list[n].flag & FLAG_SPLIT) split = 1;
        }
    } */

    // Bit-flip image
    if(flip)
    {
        uint8 temp;
        unsigned int count;

        for(count = 0; count < size; count += 1)
        {
            temp = buf[count];
            buf[count] = bitflip[temp];
        }
    }

    /* Always split 384K images */
    if(size == 0x60000)
    {
        gp_str_func.memcpy(rom + 0x00000, buf + 0x00000, 0x40000);
        gp_str_func.memcpy(rom + 0x80000, buf + 0x40000, 0x20000);
    }
    else /* Split 512K images if requested */
    if(split && (size == 0x80000))
    {
        gp_str_func.memcpy(rom + 0x00000, buf + 0x00000, 0x40000);
        gp_str_func.memcpy(rom + 0x80000, buf + 0x40000, 0x40000);
    }
    else
    {
        gp_str_func.memcpy(rom, buf, (size > 0x100000) ? 0x100000 : size);
    }

    /* Free allocated memory and exit */
    gp_mem_func.free(ptr);

/*#ifdef DOS
    // I need Allegro to handle this...
    strcpy(game_name, get_filename(game_name));
#endif */


#ifdef zzz
    #include "bitflip.h"
    uint8 header[0x200];
    uint8 *ptr = NULL, *buf = NULL;
    uint32 crc;
    unsigned long size, n, n_dummy;

	ERR_CODE err_code;
	F_HANDLE h_rfile;

    /* Default */
	gp_str_func.strcpy(game_name, filename);
	textout(filename);

	err_code = GpRelativePathSet("gp:\\gpmm\\pcegp");
	if (err_code != SM_OK)
	{
		textout((char*)"Path set failed");
	}
	err_code = GpFileGetSize(filename, &size);
	if (err_code != SM_OK)
	{
		textout((char*)"Get Size failed");
	}
	err_code = GpFileOpen(filename, OPEN_R, &h_rfile);
	if (err_code != SM_OK)
	{
		textout((char*)"File Open Failed");
	}

	buf = gp_mem_func.malloc(size);
	err_code = GpFileRead(h_rfile, buf, size, (ulong*)&n_dummy);
	if (err_code != SM_OK)
	{
		textout((char*)"File Read Failed");
		GpThreadSleep(500);
	}

	GpFileClose(h_rfile);
#endif

#ifdef Dos
    gzFile *gd = NULL;

    /* Open file */
    gd = gzopen(filename, "rb");
    if(!gd) return (0);

    /* Get file size */
    size = gzsize(gd);

    /* Allocate file data buffer */

    buf = malloc(size);
    if(!buf)
	{
		gzclose(gd);
		return (0);
	}
	gzread(gd, buf, size); // Read file data
	gzclose(gd);	// Close file
// #endif

    /* Check for 512-byte header */
    ptr = buf;
    if((size / 512) & 1)
    {
        gp_str_func.memcpy(header, buf, 0x200);
        size -= 0x200;
        buf += 0x200;
    }

    /* Generate CRC and print information */
	#ifdef Dos
    crc = crc32(0, buf, size);
	#endif
	crc = 0;

    /* Look up game CRC in the CRC database, and set up flip and
       split options accordingly */
    for(n = 0; n < (sizeof(pcecrc_list) / sizeof(t_pcecrc)); n += 1)
    {
        if(crc == pcecrc_list[n].crc)
        {
            if(pcecrc_list[n].flag & FLAG_BITFLIP) flip = 1;
            if(pcecrc_list[n].flag & FLAG_SPLIT) split = 1;
        }
    }

    /* Bit-flip image */
    if(flip)
    {
        uint8 temp;
        int count;

        for(count = 0; count < size; count += 1)
        {
            temp = buf[count];
            buf[count] = bitflip[temp];
        }
    }

    /* Always split 384K images */
    if(size == 0x60000)
    {
        gp_str_func.memcpy(rom + 0x00000, buf + 0x00000, 0x40000);
        gp_str_func.memcpy(rom + 0x80000, buf + 0x40000, 0x20000);
    }
    else /* Split 512K images if requested */
    if(split && (size == 0x80000))
    {
        gp_str_func.memcpy(rom + 0x00000, buf + 0x00000, 0x40000);
        gp_str_func.memcpy(rom + 0x80000, buf + 0x40000, 0x40000);
    }
    else
    {
        gp_str_func.memcpy(rom, buf, (size > 0x100000) ? 0x100000 : size);
    }

    /* Free allocated memory and exit */
    gp_mem_func.free(ptr);

// #ifdef DOS
    /* I need Allegro to handle this... */
    strcpy(game_name, get_filename(game_name));
#endif

    return 1;
}


int file_exist(char *filename)
{
#ifdef Dos
    FILE *fd = fopen(filename, "rb");
    if(!fd) return (0);
    fclose(fd);
    return (1);
#endif
	return 0;
}


int load_file(char *filename, char *buf, int size)
{
#ifdef Dos
    FILE *fd = fopen(filename, "rb");
    if(!fd) return (0);
    fread(buf, size, 1, fd);
    fclose(fd);
    return (1);
#endif
	return 0;
}


int save_file(char *filename, char *buf, int size)
{
#ifdef Dos
    FILE *fd = NULL;
    if(!(fd = fopen(filename, "wb"))) return (0);
    fwrite(buf, size, 1, fd);
    fclose(fd);
    return (1);
#endif
	return 0;
}